@@ -24,15 +24,21 @@ import (
 const (
 	flagAll             flagName = "all"
 	flagAllErrors       flagName = "all-errors"
+	flagAllMajor        flagName = "all-major"
+	flagAllVersions     flagName = "all-versions"
 	flagCheck           flagName = "check"
 	flagDiff            flagName = "diff"
 	flagDryRun          flagName = "dry-run"
 	flagEscape          flagName = "escape"
+	flagExact           flagName = "exact"
 	flagExpression      flagName = "expression"
 	flagExt             flagName = "ext"
 	flagFiles           flagName = "files"
 	flagForce           flagName = "force"
+	flagFrom            flagName = "from"
 	flagGlob            flagName = "name"
+	flagGroup           flagName = "group"
+	flagIdent           flagName = "ident"
 	flagIgnore          flagName = "ignore"
 	flagInject          flagName = "inject"
 	flagInjectVars      flagName = "inject-vars"
@@ -41,6 +47,8 @@ const (
 	flagLanguageVersion flagName = "language-version"
 	flagList            flagName = "list"
 	flagMerge           flagName = "merge"
+	flagMod             flagName = "mod"
+	flagNoDeps          flagName = "no-deps"
 	flagOut             flagName = "out"
 	flagOutFile         flagName = "outfile"
 	flagPackage         flagName = "package"
@@ -52,7 +60,9 @@ const (
 	flagSimplify        flagName = "simplify"
 	flagSource          flagName = "source"
 	flagStrict          flagName = "strict"
+	flagTo              flagName = "to"
 	flagTrace           flagName = "trace"
+	flagUpdateIdent     flagName = "update-ident"
 	flagVerbose         flagName = "verbose"
 	flagWithContext     flagName = "with-context"
 
@@ -82,10 +92,9 @@ func addGlobalFlags(f *pflag.FlagSet) {
 		"print information about progress")
 	f.BoolP(string(flagAllErrors), "E", false, "print all available errors")
 
-	// Deprecated flags are hidden but still work for now.
-	// TODO(mvdan): make this flag give a warning or error in early 2025.
-	f.Bool(string(flagStrict), false,
-		"report errors for lossy mappings (deprecated: use \"jsonschema+strict:\" filetype instead; see cue help filetypes)")
+	// As of v0.14, this deprecated flag now fails when used.
+	// TODO(mvdan): remove entirely in v0.16; no users should remain by then.
+	f.Bool(string(flagStrict), false, "")
 	f.MarkHidden(string(flagStrict))
 
 	f.String(string(flagCpuProfile), "", "write a CPU profile to the specified file before exiting")
@@ -103,7 +112,7 @@ func addOrphanFlags(f *pflag.FlagSet) {
 	f.Bool(string(flagWithContext), false, "import as object with contextual data")
 	f.StringArrayP(string(flagProtoPath), "I", nil, "paths in which to search for imports")
 	f.String(string(flagProtoEnum), "int", "mode for rendering enums (int|json)")
-	f.StringP(string(flagGlob), "n", "", "glob filter for non-CUE file names in directories")
+	f.StringP(string(flagGlob), "n", "", "regexp filter for non-CUE file names in directories")
 	f.Bool(string(flagMerge), true, "merge non-CUE files")
 }
 
@@ -130,6 +139,19 @@ func (f flagName) ensureAdded(cmd *Command) {
 	}
 }
 
+// IsSet reports whether f was provided by the user, which is useful to tell
+// `cue` apart from `cue --flag=` when the flag's default value is the zero value.
+func (f flagName) IsSet(cmd *Command) bool {
+	f.ensureAdded(cmd)
+	found := false
+	cmd.Flags().Visit(func(pf *pflag.Flag) {
+		if pf.Name == string(f) {
+			found = true
+		}
+	})
+	return found
+}
+
 func (f flagName) Bool(cmd *Command) bool {
 	f.ensureAdded(cmd)
 	v, _ := cmd.Flags().GetBool(string(f))
