@@ -15,19 +15,26 @@
 package toml_test
 
 import (
+	"bytes"
 	"encoding/json"
 	"io"
+	"path"
+	"reflect"
 	"strings"
 	"testing"
 
 	"github.com/go-quicktest/qt"
 	gotoml "github.com/pelletier/go-toml/v2"
 
+	"cuelang.org/go/cue/ast"
 	"cuelang.org/go/cue/ast/astutil"
 	"cuelang.org/go/cue/cuecontext"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/format"
+	"cuelang.org/go/cue/token"
 	"cuelang.org/go/encoding/toml"
+	"cuelang.org/go/internal/astinternal"
+	"cuelang.org/go/internal/cuetxtar"
 )
 
 func TestDecoder(t *testing.T) {
@@ -309,6 +316,40 @@ line two.\
 			positive: true
 			negative: false
 			`,
+	}, {
+		name: "DateTimes",
+		input: `
+			offsetDateTime1 = 1979-05-27T07:32:00Z
+			offsetDateTime2 = 1979-05-27T00:32:00-07:00
+			offsetDateTime3 = 1979-05-27T00:32:00.999999-07:00
+			localDateTime1 = 1979-05-27T07:32:00
+			localDateTime2 = 1979-05-27T00:32:00.999999
+			localDate1 = 1979-05-27
+			localTime1 = 07:32:00
+			localTime2 = 00:32:00.999999
+
+			inlineArray = [1979-05-27, 07:32:00]
+
+			notActuallyDate = "1979-05-27"
+			notActuallyTime = "07:32:00"
+			inlineArrayNotActually = ["1979-05-27", "07:32:00"]
+			`,
+		wantCUE: `
+			import "time"
+
+			offsetDateTime1: "1979-05-27T07:32:00Z" & time.Format(time.RFC3339)
+			offsetDateTime2: "1979-05-27T00:32:00-07:00" & time.Format(time.RFC3339)
+			offsetDateTime3: "1979-05-27T00:32:00.999999-07:00" & time.Format(time.RFC3339)
+			localDateTime1: "1979-05-27T07:32:00" & time.Format("2006-01-02T15:04:05")
+			localDateTime2: "1979-05-27T00:32:00.999999" & time.Format("2006-01-02T15:04:05")
+			localDate1: "1979-05-27" & time.Format(time.RFC3339Date)
+			localTime1: "07:32:00" & time.Format("15:04:05")
+			localTime2: "00:32:00.999999" & time.Format("15:04:05")
+			inlineArray: ["1979-05-27" & time.Format(time.RFC3339Date), "07:32:00" & time.Format("15:04:05")]
+			notActuallyDate: "1979-05-27"
+			notActuallyTime: "07:32:00"
+			inlineArrayNotActually: ["1979-05-27", "07:32:00"]
+			`,
 	}, {
 		name: "Arrays",
 		input: `
@@ -803,8 +844,6 @@ line two.\
 
 			// Ensure that the CUE node can be compiled into a cue.Value and validated.
 			ctx := cuecontext.New()
-			// TODO(mvdan): cue.Context can only build ast.Expr or ast.File, not ast.Node;
-			// it's then likely not the right choice for the interface to return ast.Node.
 			val := ctx.BuildFile(file)
 			qt.Assert(t, qt.IsNil(val.Err()))
 			qt.Assert(t, qt.IsNil(val.Validate()))
@@ -828,6 +867,10 @@ line two.\
 			// Ensure that the decoded CUE can be re-encoded as TOML,
 			// and the resulting TOML is still JSON-equivalent.
 			t.Run("reencode", func(t *testing.T) {
+				switch test.name {
+				case "DateTimes":
+					t.Skip("TODO(mvdan): dates and times always encode as TOML strings today")
+				}
 				sb := new(strings.Builder)
 				enc := toml.NewEncoder(sb)
 
@@ -839,7 +882,8 @@ line two.\
 				var unmarshalCueTOML any
 				err = gotoml.Unmarshal([]byte(cueTOML), &unmarshalCueTOML)
 				qt.Assert(t, qt.IsNil(err))
-				qt.Assert(t, qt.JSONEquals(jsonCUE, unmarshalCueTOML))
+
+				qt.Assert(t, qt.CmpEquals(unmarshalCueTOML, unmarshalTOML))
 			})
 		})
 	}
@@ -860,3 +904,36 @@ func unindentMultiline(s string) string {
 	s = strings.TrimSuffix(s, "\n")
 	return s
 }
+
+var (
+	typNode = reflect.TypeFor[ast.Node]()
+	typPos  = reflect.TypeFor[token.Pos]()
+)
+
+func TestDecoderTxtar(t *testing.T) {
+	test := cuetxtar.TxTarTest{
+		Root: "testdata",
+		Name: "decode",
+	}
+
+	test.Run(t, func(t *cuetxtar.Test) {
+		for _, file := range t.Archive.Files {
+			if strings.HasPrefix(file.Name, "out/") {
+				continue
+			}
+			dec := toml.NewDecoder(file.Name, bytes.NewReader(file.Data))
+			node, err := dec.Decode()
+			qt.Assert(t, qt.IsNil(err))
+
+			// Show all valid node positions.
+			out := astinternal.AppendDebug(nil, node, astinternal.DebugConfig{
+				OmitEmpty: true,
+				Filter: func(v reflect.Value) bool {
+					t := v.Type()
+					return t.Implements(typNode) || t.Kind() == reflect.Slice || t == typPos
+				},
+			})
+			t.Writer(path.Join(file.Name, "positions")).Write(out)
+		}
+	})
+}
